// Copyright (c) FIRST and other WPILib contributors.
// Open Source Software; you can modify and/or share it under the terms of
// the WPILib BSD license file in the root directory of this project.

#include "cscore_cpp.h"

namespace cs {

std::vector<std::string> GetNetworkInterfaces() {
  return std::vector<std::string>{};  // TODO
}

}  // namespace cs
