// Copyright (c) FIRST and other WPILib contributors.
// Open Source Software; you can modify and/or share it under the terms of
// the WPILib BSD license file in the root directory of this project.

#ifndef WPIUTIL_WPI_MIMETYPES_H_
#define WPIUTIL_WPI_MIMETYPES_H_

#include "wpi/StringRef.h"

namespace wpi {

StringRef MimeTypeFromPath(StringRef path);

}  // namespace wpi

#endif  // WPIUTIL_WPI_MIMETYPES_H_
