// Copyright (c) FIRST and other WPILib contributors.
// Open Source Software; you can modify and/or share it under the terms of
// the WPILib BSD license file in the root directory of this project.

#ifndef CSCORE_CSCORE_H_
#define CSCORE_CSCORE_H_

/* C API */
#include "cscore_c.h"

#ifdef __cplusplus
/* C++ API */
#include "cscore_cpp.h"
#include "cscore_oo.h"
#endif /* __cplusplus */

#endif  // CSCORE_CSCORE_H_
