// Copyright (c) FIRST and other WPILib contributors.
// Open Source Software; you can modify and/or share it under the terms of
// the WPILib BSD license file in the root directory of this project.

#ifndef WPIUTIL_WPI_DEPRECATED_H_
#define WPIUTIL_WPI_DEPRECATED_H_

#ifndef WPI_DEPRECATED
#define WPI_DEPRECATED(msg) [[deprecated(msg)]]
#endif

#endif  // WPIUTIL_WPI_DEPRECATED_H_
